import os
import sys

from pyspark.sql import SparkSession

from pyspark_explorer import ui
from pyspark_explorer.explorer import Explorer


def run() -> None:
    # ensure no spark errors if firewall restrictions exists
    if os.getenv("SPARK_LOCAL_IP") is None:
        os.environ["SPARK_LOCAL_IP"] = "127.0.0.1"

    spark = (SparkSession.builder
             .master("local[2]")
             # ERRORS will be briefly displayed on screen if log4j.properties file does not exist or does not forward logs to a file
             .config("spark.log.level", "ERROR")
             .config("spark.driver.extraJavaOptions", "-Dlog4j.configuration=file:log4j.properties")
             .appName("pyspark_explorer")
             .getOrCreate())

    explorer = Explorer(spark, sys.argv[1] if len(sys.argv)>1 else "/")
    app = ui.DataApp(explorer)
    app.run()

    spark.stop()



import datetime
import os

from pyspark.sql import SparkSession
from pyspark.sql.types import StructType, StructField, StringType, IntegerType, DateType, ArrayType

from pyspark_explorer.data_table import DataFrameTable


class TestSparkDataTable:
    @classmethod
    def setup_class(cls):
        cls.spark = SparkSession.builder.appName("pyspark_test01").getOrCreate()

    @classmethod
    def teardown_class(cls):
        cls.spark.stop()

    def test_simple_fields(self) -> None:
        schema = StructType([
            StructField("id", IntegerType(), True),
            StructField("text", StringType(), True),
            StructField("date", DateType(), True),
        ])
        current_dir = os.path.dirname(__file__)
        df = TestSparkDataTable.spark.read.schema(schema).csv(path=f"file:///{current_dir}/spark_test01.csv")

        df_cols = df.schema.fields
        df_rows = df.take(2)
        tab = DataFrameTable(df_cols, df_rows)

        expected_cols = [
            {'col_index': 0, 'name': 'id', "kind": "simple", 'type': 'IntegerType', 'field_type': IntegerType()},
            {'col_index': 1, 'name': 'text', "kind": "simple", 'type': 'StringType', 'field_type': StringType()},
            {'col_index': 2, 'name': 'date', "kind": "simple", 'type': 'DateType', 'field_type': DateType()}
        ]
        expected_rows = [
            {'row_index': 0, 'row': [
                {'display_value': '1', 'value': 1},
                {'display_value': 'abc', 'value': 'abc'},
                {'display_value': '2024-01-01', 'value': datetime.date(2024, 1, 1)}]},
            {'row_index': 1, 'row': [
                {'display_value': '2', 'value': 2},
                {'display_value': 'def', 'value': 'def'},
                {'display_value': '2024-01-02', 'value': datetime.date(2024, 1, 2)}]}
        ]

        assert tab.columns == expected_cols
        assert tab.rows == expected_rows

    def test_embedded_array(self) -> None:
        schema = StructType([StructField("id", IntegerType()),
                             StructField("info", ArrayType(StructType([
                                 StructField("num", IntegerType(), True),
                                 StructField("text", StringType(), True)])))])

        current_dir = os.path.dirname(__file__)
        df = TestSparkDataTable.spark.read.schema(schema).json(path=f"file:///{current_dir}/spark_test02.json")

        df_cols = df.schema.fields
        df_rows = df.take(2)
        tab = DataFrameTable(df_cols, df_rows)

        expected_cols = [
            {'col_index': 0, 'name': 'id', "kind": "simple", 'type': 'IntegerType', 'field_type': IntegerType()},
            {'col_index': 1, 'name': 'info', "kind": "array", 'type': 'ArrayType', 'field_type': StructType([
                StructField('num', IntegerType(), True),
                StructField('text', StringType(), True)])}]

        expected_rows = [
            {'row_index': 0,'row': [
                {'display_value': '1', 'value': 1},
                {'display_value': "(num=101, text='aaa'), (num=201, text='bbb')",
                'value': [{'row_index': 0, 'row': [{'display_value': "num=101, text='aaa'",
                    'value':
                        {'row_index': 0, 'row': [
                            {'display_value': '101', 'value': 101},
                            {'display_value': 'aaa', 'value': 'aaa'}]}}]},
                        {'row_index': 1, 'row': [
                            {'display_value': "num=201, text='bbb'",
                            'value':
                                {'row_index': 0, 'row': [
                                    {'display_value': '201', 'value': 201},
                                    {'display_value': 'bbb', 'value': 'bbb'}]}}]}]}]},

            {'row_index': 1,'row': [
                {'display_value': '2', 'value': 2},
                {'display_value': "(num=102, text='ccc'), (num=202, text='ddd')",
                'value': [{'row_index': 0, 'row': [
                    {'display_value': "num=102, text='ccc'",
                    'value':
                        {'row_index': 0, 'row': [
                            {'display_value': '102', 'value': 102},
                            {'display_value': 'ccc', 'value': 'ccc'}]}}]},
                        {'row_index': 1, 'row': [
                            {'display_value': "num=202, text='ddd'",
                            'value':
                                {'row_index': 0, 'row': [
                                    {'display_value': '202', 'value': 202},
                                    {'display_value': 'ddd', 'value': 'ddd'}]}}]}]}]}]

        print(tab.columns)
        print(tab.rows)

        assert tab.columns == expected_cols
        assert tab.rows == expected_rows


from pyspark.sql.types import StructType, StructField, StringType, Row, DateType, LongType, IntegerType, ArrayType

from pyspark_explorer.data_table import DataFrameTable, extract_embedded_table


class TestDataTable:
    @staticmethod
    def __prepare_simple_text_field__(val1: str, val2: str) -> ([StructField], [Row], {}, {}):
        schema = [StructField("text", StringType())]
        rows = [Row(text=val1), Row(text=val2)]
        expected_cols = [{"col_index": 0, "name": "text", "kind": "simple", "type": "StringType", "field_type": schema[0].dataType}]
        expected_rows = [
            {"row_index": 0, "row": [{"value": val1, "display_value": val1}]},
            {"row_index": 1, "row": [{"value": val2, "display_value": val2}]},
        ]
        return schema, rows, expected_cols, expected_rows

    @staticmethod
    def __prepare_simple_num_field__(val1: int, val2: int) -> ([StructField], [Row], {}, {}):
        schema = [StructField("nums", IntegerType())]
        rows = [Row(nums=val1),Row(nums=val2)]
        expected_cols = [
            {"col_index": 0, "name": "nums", "kind": "simple", "type": "IntegerType", "field_type": schema[0].dataType},
        ]
        expected_rows = [
            {"row_index": 0, "row": [{"value": val1, "display_value": DataFrameTable.disp_value(val1)}]},
            {"row_index": 1, "row": [{"value": val2, "display_value": DataFrameTable.disp_value(val2)}]},
        ]
        return schema, rows, expected_cols, expected_rows


    @staticmethod
    def __prepare_multiple_simple_fields_schema__() -> ([StructField], []):
        schema = [StructField("id", LongType()), StructField("text", StringType()), StructField("date", DateType())]
        expected_cols = [
            {"col_index": 0, "name": "id", "kind": "simple", "type": "LongType", "field_type": schema[0].dataType},
            {"col_index": 1, "name": "text", "kind": "simple", "type": "StringType", "field_type": schema[1].dataType},
            {"col_index": 2, "name": "date", "kind": "simple", "type": "DateType", "field_type": schema[2].dataType}
        ]
        return schema, expected_cols


    @staticmethod
    def __prepare_multiple_simple_fields_row__(val: [], index: int) -> (Row, {}):
        row = Row(id=val[0], text=val[1], date=val[2])
        _, expected_cols = TestDataTable.__prepare_multiple_simple_fields_schema__()
        expected_row = {"row_index": index, "row": [
            {"value": val[0], "display_value": DataFrameTable.disp_value(val[0])},
            {"value": val[1], "display_value": val[1]},
            {"value": val[2], "display_value": val[2]}
        ]}
        return row, expected_row


    @staticmethod
    def __prepare_multiple_simple_fields__(val1: [], val2: []) -> ([StructField], [Row], {}, {}):
        schema, expected_cols = TestDataTable.__prepare_multiple_simple_fields_schema__()
        row1, expected_row1 = TestDataTable.__prepare_multiple_simple_fields_row__(val1, 0)
        row2, expected_row2 = TestDataTable.__prepare_multiple_simple_fields_row__(val2, 1)
        rows = [row1, row2]
        expected_rows = [expected_row1, expected_row2]
        return schema, rows, expected_cols, expected_rows


    def test_one_simple_field(self) -> None:
        schema, rows, expected_cols, expected_rows = TestDataTable.__prepare_simple_text_field__("some text 1", "some text 2")
        tab = DataFrameTable(schema, rows)
        assert tab.columns == expected_cols
        assert tab.rows == expected_rows
        assert tab.column_names == [schema[0].name]
        assert tab.row_values == [["some text 1"],["some text 2"]]


    def test_multiple_simple_fields(self) -> None:
        schema, rows, expected_cols, expected_rows = TestDataTable.__prepare_multiple_simple_fields__(
             [100, "some text 1", "2024-01-01"],[101, "some text 2", "2024-01-02"])
        tab = DataFrameTable(schema, rows)
        assert tab.columns == expected_cols
        assert tab.rows == expected_rows
        assert tab.column_names == [schema[0].name, schema[1].name, schema[2].name]
        assert tab.row_values == [["100", "some text 1", "2024-01-01"],["101", "some text 2", "2024-01-02"]]


    def test_array_of_single_field(self) -> None:
        # first test internal fields containing arrays
        inner_schema1, inner_rows1, inner_expected_cols1, inner_expected_rows1 = TestDataTable.__prepare_simple_num_field__(1,2)
        inner_schema2, inner_rows2, inner_expected_cols2, inner_expected_rows2 = TestDataTable.__prepare_simple_num_field__(3,4)
        inner_tab1 = DataFrameTable(inner_schema1, inner_rows1)
        inner_tab2 = DataFrameTable(inner_schema2, inner_rows2)
        assert inner_tab1.columns == inner_expected_cols1
        assert inner_tab2.columns == inner_expected_cols2
        assert inner_tab1.rows == inner_expected_rows1
        assert inner_tab2.rows == inner_expected_rows2

        # now test complex schema with embedded array of a simple field
        schema = [StructField("nums", ArrayType(IntegerType()))]
        rows = [Row(num=[1,2]),Row(num=[3,4])]
        tab = DataFrameTable(schema, rows)

        expected_cols = [
            {"col_index": 0, "name": "nums", "kind": "array", "type": "ArrayType", "field_type": schema[0].dataType.elementType},
        ]
        assert tab.columns == expected_cols
        assert tab.column_names == ["nums"]
        expected_rows = [
            {"row_index": 0, "row": [{"value": inner_expected_rows1, "display_value": DataFrameTable.disp_value([1,2])}]},
            {"row_index": 1, "row": [{"value": inner_expected_rows2, "display_value": DataFrameTable.disp_value([3,4])}]},
        ]
        assert tab.rows == expected_rows
        assert tab.row_values == [[str([1,2])],[str([3,4])]]


    def test_empty_array_of_single_field(self) -> None:
        # now test complex schema with embedded empty array of a simple field
        schema = [StructField("nums", ArrayType(IntegerType()))]
        rows = None
        tab = DataFrameTable(schema, rows)

        expected_cols = [
            {"col_index": 0, "name": "nums", "kind": "array", "type": "ArrayType", "field_type": schema[0].dataType.elementType},
        ]
        assert tab.columns == expected_cols
        assert tab.column_names == ["nums"]
        expected_rows = []
        assert tab.rows == expected_rows
        assert tab.row_values == []


    def test_embedded_struct_field(self) -> None:
        # first test internal fields (struct fields)
        inner_row1, inner_expected_row1 = TestDataTable.__prepare_multiple_simple_fields_row__([11, "some text 1", "2024-02-01"], 0)
        inner_row2, inner_expected_row2 = TestDataTable.__prepare_multiple_simple_fields_row__([13, "some text 3", "2024-02-03"], 0)
        inner_schema, inner_expected_cols = TestDataTable.__prepare_multiple_simple_fields_schema__()

        inner_tab1 = DataFrameTable(inner_schema, [inner_row1])
        inner_tab2 = DataFrameTable(inner_schema, [inner_row2])

        assert inner_tab1.columns == inner_expected_cols
        assert inner_tab2.columns == inner_expected_cols
        assert inner_tab1.rows == [inner_expected_row1]
        assert inner_tab2.rows == [inner_expected_row2]

        # now test complex schema with embedded struct field
        schema = [StructField("row_id", IntegerType()), StructField("struct", StructType(inner_schema))]
        rows = [Row(id=1, struct=inner_row1), Row(id=2, struct=inner_row2)]
        tab = DataFrameTable(schema, rows)

        expected_cols = [
            {"col_index": 0, "name": "row_id", "kind": "simple", "type": "IntegerType", "field_type": schema[0].dataType},
            {"col_index": 1, "name": "struct", "kind": "struct", "type": "StructType", "field_type": schema[1].dataType}
        ]

        assert tab.columns == expected_cols
        assert tab.column_names == ["row_id", "struct"]

        expected_rows = [
            {"row_index": 0, "row": [
                {"value": 1, "display_value": "1"},
                {"value": inner_expected_row1, "display_value": DataFrameTable.disp_value(inner_row1)},
            ]},
            {"row_index": 1, "row": [
                {"value": 2, "display_value": "2"},
                {"value": inner_expected_row2, "display_value": DataFrameTable.disp_value(inner_row2)},
            ]},
        ]

        assert tab.rows == expected_rows
        assert tab.row_values == [
            ["1", DataFrameTable.disp_value(inner_row1)[:DataFrameTable.TEXT_LEN]],
            ["2", DataFrameTable.disp_value(inner_row2)[:DataFrameTable.TEXT_LEN]]]

        # now drill down to details and make sure the results are the same
        extracted_tab1 = extract_embedded_table(tab, 1, 0)
        assert extracted_tab1.columns == inner_expected_cols
        assert extracted_tab1.rows == [inner_expected_row1]


    def test_embedded_empty_struct_field(self) -> None:
        inner_schema, inner_expected_cols = TestDataTable.__prepare_multiple_simple_fields_schema__()

        # test complex schema with embedded empty struct field
        schema = [StructField("row_id", IntegerType()), StructField("struct", StructType(inner_schema))]
        rows = None
        tab = DataFrameTable(schema, rows)

        expected_cols = [
            {"col_index": 0, "name": "row_id", "kind": "simple", "type": "IntegerType", "field_type": schema[0].dataType},
            {"col_index": 1, "name": "struct", "kind": "struct", "type": "StructType", "field_type": schema[1].dataType}
        ]

        assert tab.columns == expected_cols
        assert tab.column_names == ["row_id", "struct"]

        expected_rows = []

        assert tab.rows == expected_rows
        assert tab.row_values == []


    @staticmethod
    def __array_to_row__(schema:[StructField], arr: []) -> [Row]:
        field_names = list(map(lambda f: f.name, schema))
        res_rows: [Row] = []
        for elem in arr:
            pairs = zip(field_names, elem)
            row_to_add = Row(**dict(pairs))
            res_rows.append(row_to_add)

        return res_rows

    def test_array_of_struct_field(self) -> None:
        # first test internal fields (struct fields)
        input_rows1 = [[11, "some text 1", "2024-02-01"], [12, "some text 2", "2024-02-02"]]
        input_rows2 = [[13, "some text 3", "2024-02-03"], [14, "some text 4", "2024-02-04"]]
        inner_schema1, inner_rows1, inner_expected_cols1, inner_expected_rows1 = TestDataTable.__prepare_multiple_simple_fields__(
            input_rows1[0], input_rows1[1])
        inner_schema2, inner_rows2, inner_expected_cols2, inner_expected_rows2 = TestDataTable.__prepare_multiple_simple_fields__(
            input_rows2[0], input_rows2[1])
        inner_tab1 = DataFrameTable(inner_schema1, inner_rows1)
        inner_tab2 = DataFrameTable(inner_schema2, inner_rows2)
        assert inner_tab1.columns == inner_expected_cols1
        assert inner_tab2.columns == inner_expected_cols2
        assert inner_tab1.rows == inner_expected_rows1
        assert inner_tab2.rows == inner_expected_rows2

        # then embed the struct into single field (2 separate tabs, each for one row in the final table)
        inner_embedded_schema1 = [StructField("structs",StructType(inner_schema1))]
        inner_embedded_schema2 = [StructField("structs",StructType(inner_schema2))]
        inner_embedded_rows1 = [Row(structs=inner_rows1[0]),Row(structs=inner_rows1[1])]
        inner_embedded_rows2 = [Row(structs=inner_rows2[0]),Row(structs=inner_rows2[1])]
        inner_embedded_tab1 = DataFrameTable(inner_embedded_schema1, inner_embedded_rows1)
        inner_embedded_tab2 = DataFrameTable(inner_embedded_schema2, inner_embedded_rows2)
        inner_embedded_expected_cols1 = [
            {"col_index": 0, "name": "structs", "kind": "struct", "type": "StructType", "field_type": inner_embedded_schema1[0].dataType}
        ]
        inner_embedded_expected_cols2 = [
            {"col_index": 0, "name": "structs", "kind": "struct", "type": "StructType", "field_type": inner_embedded_schema2[0].dataType}
        ]

        assert inner_embedded_tab1.columns == inner_embedded_expected_cols1
        assert inner_embedded_tab2.columns == inner_embedded_expected_cols2

        inner_expected_rows1_upd = inner_expected_rows1[1]
        inner_expected_rows1_upd["row_index"]=0
        inner_embedded_expected_rows1 = [
            {"row_index": 0, "row": [{"value": inner_expected_rows1[0], "display_value": DataFrameTable.disp_value(inner_rows1[0])}]},
            {"row_index": 1, "row": [{"value": inner_expected_rows1_upd, "display_value": DataFrameTable.disp_value(inner_rows1[1])}]}]
        inner_expected_rows2_upd = inner_expected_rows2[1]
        inner_expected_rows2_upd["row_index"]=0
        inner_embedded_expected_rows2 = [
            {"row_index": 0, "row": [{"value": inner_expected_rows2[0], "display_value": DataFrameTable.disp_value(inner_rows2[0])}]},
            {"row_index": 1, "row": [{"value": inner_expected_rows2_upd, "display_value": DataFrameTable.disp_value(inner_rows2[1])}]}]

        assert inner_embedded_tab1.rows == inner_embedded_expected_rows1
        assert inner_embedded_tab2.rows == inner_embedded_expected_rows2

        # now test complex schema with embedded array of struct field
        inner_rows1_as_rows = self.__array_to_row__(inner_schema1, inner_rows1)
        inner_rows2_as_rows = self.__array_to_row__(inner_schema2, inner_rows2)
        schema = [StructField("id", IntegerType()), StructField("structs", ArrayType(StructType(inner_schema1)))]
        rows = [Row(id=1, structs=inner_rows1_as_rows), Row(id=2, structs=inner_rows2_as_rows)]
        tab = DataFrameTable(schema, rows)

        expected_cols = [
            {"col_index": 0, "name": "id", "kind": "simple", "type": "IntegerType", "field_type": schema[0].dataType},
            {"col_index": 1, "name": "structs", "kind": "array", "type": "ArrayType", "field_type": schema[1].dataType.elementType}
        ]
        assert tab.columns == expected_cols
        assert tab.column_names == ["id", "structs"]

        expected_rows = [
            {"row_index": 0, "row": [
                {"value": 1, "display_value": "1"},
                {"value": inner_embedded_expected_rows1, "display_value": DataFrameTable.disp_value(inner_rows1_as_rows)},
            ]},
            {"row_index": 1, "row": [
                {"value": 2, "display_value": "2"},
                {"value": inner_embedded_expected_rows2, "display_value": DataFrameTable.disp_value(inner_rows2_as_rows)},
            ]},
        ]

        assert tab.rows == expected_rows
        assert tab.row_values == [
            ["1",DataFrameTable.disp_value(inner_rows1_as_rows)[:DataFrameTable.TEXT_LEN]],
            ["2",DataFrameTable.disp_value(inner_rows2_as_rows)[:DataFrameTable.TEXT_LEN]]]

        # now drill down to details and make sure the results are the same
        extracted_tab1 = extract_embedded_table(tab, 1, 0)
        assert extracted_tab1 is not None
        # this is the same as expected_cols[1] but with index 0
        assert extracted_tab1.columns == [{"col_index": 0, "name": "structs", "kind": "struct", "type": "StructType", "field_type": schema[1].dataType.elementType}]
        assert extracted_tab1.rows == inner_embedded_expected_rows1
        assert extracted_tab1.column_names == ["structs"]
        assert extracted_tab1.row_values == [
            [DataFrameTable.disp_value(inner_rows1[0])[:DataFrameTable.TEXT_LEN]],
            [DataFrameTable.disp_value(inner_rows1[1])[:DataFrameTable.TEXT_LEN]]]

        extracted_tab2 = extract_embedded_table(extracted_tab1, 0, 0)
        assert extracted_tab2 is not None
        assert extracted_tab2.columns == inner_expected_cols1
        assert extracted_tab2.rows == [inner_expected_rows1[0]]
        assert extracted_tab2.column_names == ["id", "text", "date"]
        # this is an input row but mapped to strings
        assert extracted_tab2.row_values == [['11', 'some text 1', '2024-02-01']]


    def test_empty_array_of_struct_field(self) -> None:
        # first test internal fields (struct fields)
        input_rows1 = [[1, "not important", "2024-01-01"],[1, "not important", "2024-01-01"]]
        inner_schema1, inner_rows1, inner_expected_cols1, inner_expected_rows1 = TestDataTable.__prepare_multiple_simple_fields__(
            input_rows1[0], input_rows1[1])
        inner_tab1 = DataFrameTable(inner_schema1, inner_rows1)
        assert inner_tab1.columns == inner_expected_cols1

        # then embed the struct into single field (2 separate tabs, each for one row in the final table)
        inner_embedded_schema1 = [StructField("structs",StructType(inner_schema1))]
        inner_embedded_rows1 = [Row(structs=inner_rows1[0])]
        inner_embedded_tab1 = DataFrameTable(inner_embedded_schema1, inner_embedded_rows1)
        inner_embedded_expected_cols1 = [
            {"col_index": 0, "name": "structs", "kind": "struct", "type": "StructType", "field_type": inner_embedded_schema1[0].dataType}
        ]

        assert inner_embedded_tab1.columns == inner_embedded_expected_cols1

        # now test complex schema with embedded array of struct field
        schema = [StructField("id", IntegerType()), StructField("structs", ArrayType(StructType(inner_schema1)))]
        rows = [Row(id=1, structs=None), Row(id=2, structs=None)]
        tab = DataFrameTable(schema, rows)

        expected_cols = [
            {"col_index": 0, "name": "id", "kind": "simple", "type": "IntegerType", "field_type": schema[0].dataType},
            {"col_index": 1, "name": "structs", "kind": "array", "type": "ArrayType", "field_type": schema[1].dataType.elementType}
        ]
        assert tab.columns == expected_cols
        assert tab.column_names == ["id", "structs"]

        expected_rows = [
            {"row_index": 0, "row": [{"value": 1, "display_value": "1"},{"value": [], "display_value": "[]"}]},
            {"row_index": 1, "row": [{"value": 2, "display_value": "2"},{"value": [], "display_value": "[]"}]},
        ]

        assert tab.rows == expected_rows
        assert tab.row_values == [["1","[]"],["2","[]"]]

        # now drill down to details and make sure the results are the same
        extracted_tab1 = extract_embedded_table(tab, 1, 0)
        assert extracted_tab1 is not None
        # this is the same as expected_cols[1] but with index 0
        assert extracted_tab1.columns == [{"col_index": 0, "name": "structs", "kind": "struct", "type": "StructType", "field_type": schema[1].dataType.elementType}]
        assert extracted_tab1.rows == []
        assert extracted_tab1.column_names == ["structs"]
        assert extracted_tab1.row_values == []


    def test_struct_field_expansion(self) -> None:
        # first test internal fields (struct fields)
        inner_row1, inner_expected_row1 = TestDataTable.__prepare_multiple_simple_fields_row__([11, "some text 1", "2024-02-01"], 0)
        inner_row2, inner_expected_row2 = TestDataTable.__prepare_multiple_simple_fields_row__([13, "some text 3", "2024-02-03"], 0)
        inner_schema, inner_expected_cols = TestDataTable.__prepare_multiple_simple_fields_schema__()

        schema = [StructField("row_id", IntegerType()), StructField("struct", StructType(inner_schema))]
        rows = [Row(id=1, struct=inner_row1), Row(id=2, struct=inner_row2)]
        tab = DataFrameTable(schema, data = rows, expand_structs = True)

        expected_cols = [
            {"col_index": 0, "name": "row_id", "kind": "simple", "type": "IntegerType", "field_type": schema[0].dataType},
            {"col_index": 1, "name": "struct", "kind": "struct", "type": "StructType", "field_type": schema[1].dataType},
            {"col_index": 2, "name": "*id", "kind": "simple", "type": "LongType", "field_type": inner_schema[0].dataType},
            {"col_index": 3, "name": "*text", "kind": "simple", "type": "StringType", "field_type": inner_schema[1].dataType},
            {"col_index": 4, "name": "*date", "kind": "simple", "type": "DateType", "field_type": inner_schema[2].dataType}
        ]

        assert tab.columns == expected_cols
        assert tab.column_names == ["row_id", "struct", "*id", "*text", "*date"]

        expected_rows = [
            {"row_index": 0, "row": [
                {"value": 1, "display_value": "1"},
                {"value": inner_expected_row1, "display_value": DataFrameTable.disp_value(inner_row1)},
                {"value": inner_expected_row1["row"][0]["value"], "display_value": DataFrameTable.disp_value(inner_expected_row1["row"][0]["value"])},
                {"value": inner_expected_row1["row"][1]["value"], "display_value": DataFrameTable.disp_value(inner_expected_row1["row"][1]["value"])},
                {"value": inner_expected_row1["row"][2]["value"], "display_value": DataFrameTable.disp_value(inner_expected_row1["row"][2]["value"])},
            ]},
            {"row_index": 1, "row": [
                {"value": 2, "display_value": "2"},
                {"value": inner_expected_row2, "display_value": DataFrameTable.disp_value(inner_row2)},
                {"value": inner_expected_row2["row"][0]["value"],"display_value": DataFrameTable.disp_value(inner_expected_row2["row"][0]["value"])},
                {"value": inner_expected_row2["row"][1]["value"],"display_value": DataFrameTable.disp_value(inner_expected_row2["row"][1]["value"])},
                {"value": inner_expected_row2["row"][2]["value"],"display_value": DataFrameTable.disp_value(inner_expected_row2["row"][2]["value"])},
            ]},
        ]

        assert tab.rows == expected_rows
        assert tab.row_values == [
            ["1", DataFrameTable.disp_value(inner_row1)[:DataFrameTable.TEXT_LEN], *[str(v["value"]) for v in inner_expected_row1["row"]]],
            ["2", DataFrameTable.disp_value(inner_row2)[:DataFrameTable.TEXT_LEN], *[str(v["value"]) for v in inner_expected_row2["row"]]]]


    @staticmethod
    def __double_embedded_struct__() -> [StructField]:
        return [
            StructField("row_id1", IntegerType()),
            StructField("struct1", ArrayType(StructType([
                StructField("row_id2", IntegerType()),
                StructField("struct2", ArrayType(StructType([
                    StructField("row_id3", IntegerType()),
                    StructField("row_value3", IntegerType())
                ])))])))]


    def test_embedded_array_of_struct_field_expansion(self) -> None:
        # this test is quite long but this form should be readable enough
        schema = self.__double_embedded_struct__()
        rows = [Row(
            row_id1=1,
            struct1 = [Row(
                row_id2=11,
                struct2 = [Row(
                    row_id3=111, row_value3=911
                )])])]

        tab = DataFrameTable(schema, data = rows, expand_structs = True)

        expected_cols = [
            {'col_index': 0, 'name': 'row_id1', 'kind': 'simple', 'type': 'IntegerType', 'field_type': IntegerType()},
            {'col_index': 1, 'name': 'struct1', 'kind': 'array', 'type': 'ArrayType', 'field_type': schema[1].dataType.elementType}
        ]

        expected_rows = [
            {"row_index": 0, "row": [
                {"display_value": "1", "value": 1},
                {"display_value": DataFrameTable.disp_value(rows[0].struct1), "value": [
                    {"row_index": 0, "row": [
                        {"display_value": DataFrameTable.disp_value(rows[0].struct1[0]), "value":
                            {"row_index": 0, "row": [
                                {"display_value": "11", "value": 11},
                                {"display_value": DataFrameTable.disp_value(rows[0].struct1[0].struct2), "value": [
                                    {"row_index": 0, "row": [
                                        {"display_value": DataFrameTable.disp_value(rows[0].struct1[0].struct2[0]), "value":
                                            {"row_index": 0, "row": [
                                                {"display_value": "111", "value": 111},
                                                {"display_value": "911", "value": 911}
                                            ]}
                                        }]
                                     }]
                                }]
                            }
                         }]
                    }]
                }]
             }]

        assert tab.columns==expected_cols
        assert tab.rows == expected_rows

        # now extract embedded table for struct1
        tab2 = extract_embedded_table(tab, 1,0,expand_structs = True)

        expected_cols2 = [
            {'col_index': 0, 'name': 'struct1', 'kind': 'struct', 'type': 'StructType', 'field_type': schema[1].dataType.elementType},
            {'col_index': 1, 'name': '*row_id2', 'kind': 'simple', 'type': 'IntegerType', 'field_type': IntegerType()},
            {'col_index': 2, 'name': '*struct2', 'kind': 'array', 'type': 'ArrayType',
             # here we drill down from array to array contents (elementType)
             'field_type': schema[1].dataType.elementType[1].dataType.elementType}
        ]

        assert tab2.columns==expected_cols2

        expected_rows2 = [
            {"row_index": 0, "row": [
                {"display_value": DataFrameTable.disp_value(rows[0].struct1[0]), "value":
                    {"row_index": 0, "row": [
                        {"display_value": "11", "value": 11},
                        {"display_value": DataFrameTable.disp_value(rows[0].struct1[0].struct2), "value": [
                            {"row_index": 0, "row": [
                                {"display_value": DataFrameTable.disp_value(rows[0].struct1[0].struct2[0]), "value":
                                    {"row_index": 0, "row": [
                                        {"display_value": "111", "value": 111},
                                        {"display_value": "911", "value": 911}
                                    ]}
                                 }]
                             }]
                         }]
                     }
                 },
                {"display_value": "11", "value": 11},
                {"display_value": DataFrameTable.disp_value(rows[0].struct1[0].struct2), "value": [
                    {"row_index": 0, "row": [
                        {"display_value": "row_id3=111, row_value3=911", "value":
                            {"row_index": 0, "row": [
                                {"display_value": "111", "value": 111},
                                {"display_value": "911", "value": 911}
                            ]}
                         }
                    ]}
                ]}]
            }]

        assert tab2.rows==expected_rows2

        # now extract embedded table for struct2
        tab3 = extract_embedded_table(tab2, 0,0,expand_structs = True)

        expected_cols3 = [
            {'col_index': 0, 'name': 'row_id2', 'kind': 'simple', 'type': 'IntegerType', 'field_type': IntegerType()},
            {'col_index': 1, 'name': 'struct2', 'kind': 'array', 'type': 'ArrayType',
             # here we drill down from array to array contents (elementType)
             'field_type': schema[1].dataType.elementType[1].dataType.elementType}
        ]

        assert tab3.columns==expected_cols3

        expected_rows3 = [
            {"row_index": 0, "row": [
                {"display_value": "11", "value": 11},
                {"display_value": DataFrameTable.disp_value(rows[0].struct1[0].struct2), "value": [
                    {"row_index": 0, "row": [
                        {"display_value": DataFrameTable.disp_value(rows[0].struct1[0].struct2[0]), "value":
                            {"row_index": 0, "row": [
                                {"display_value": "111", "value": 111},
                                {"display_value": "911", "value": 911}
                            ]}
                         },
                    ]}
                ]}
            ]}
        ]

        assert tab3.rows==expected_rows3

    def test_generate_structure_as_tree_with_embedded_array(self) -> None:
        schema = self.__double_embedded_struct__()

        tab = DataFrameTable(schema = schema, data=[], expand_structs = False)

        expected_tree = [
            {"name": "row_id1", "kind": "simple", "type": "IntegerType", "subfields": []},
            {"name": "struct1", "kind": "array", "type": "ArrayType", "subfields": [
                {"name": "row_id2", "kind": "simple", "type": "IntegerType", "subfields": []},
                {"name": "struct2", "kind": "array", "type": "ArrayType", "subfields": [
                    {"name": "row_id3", "kind": "simple", "type": "IntegerType", "subfields": []},
                    {"name": "row_value3", "kind": "simple", "type": "IntegerType", "subfields": []}
                ]}
            ]}
        ]

        assert tab.schema_tree == expected_tree


    def test_generate_structure_as_tree_with_embedded_struct(self) -> None:
        schema = [
            StructField("row_id1", IntegerType()),
            StructField("struct1", ArrayType(StructType([
                StructField("row_id2", IntegerType()),
                StructField("struct2", StructType([
                    StructField("row_id3", IntegerType()),
                    StructField("row_value3", IntegerType())
                ]))])))]

        tab = DataFrameTable(schema = schema, data=[], expand_structs = False)

        expected_tree = [
            {"name": "row_id1", "kind": "simple", "type": "IntegerType", "subfields": []},
            {"name": "struct1", "kind": "array", "type": "ArrayType", "subfields": [
                {"name": "row_id2", "kind": "simple", "type": "IntegerType", "subfields": []},
                {"name": "struct2", "kind": "struct", "type": "StructType", "subfields": [
                    {"name": "row_id3", "kind": "simple", "type": "IntegerType", "subfields": []},
                    {"name": "row_value3", "kind": "simple", "type": "IntegerType", "subfields": []}
                ]}
            ]}
        ]

        assert tab.schema_tree == expected_tree



from typing import Any

from pyspark.sql.types import StructField, Row, StructType, ArrayType, DataType
import copy


class DataFrameTable:
    TEXT_LEN = 50

    # allow original rows (Row type) or previously transformed rows (when drilling to details)
    def __init__(self, schema: [StructField], data=None, transformed_data=None, expand_structs: bool = False):
        if transformed_data is None:
            transformed_data = []
        if data is None:
            data = []
        self._schema: [StructField] = schema
        self._data: [Row] = data
        self._expand_structs: bool = expand_structs

        self.columns = []
        self.column_names = []
        self.schema_tree = []
        self.rows = []
        self.row_values = []
        self.__extract_columns__()

        if len(data):
            self.__extract_rows__()
        else:
            self.__set_rows__(transformed_data)

        if self._expand_structs:
            self.__expand_structs__()


    @staticmethod
    def __extract_kind__(field: StructField) -> str:
        if type(field.dataType) == StructType:
            kind = "struct"
        elif type(field.dataType) == ArrayType:
            kind = "array"
        else:
            kind = "simple"
        return kind


    def __expand_structs__(self):
        new_cols = [] # NOTE: we cannot modify self.columns on the fly in the loop below, we would modify the loop
        new_rows = copy.deepcopy(self.rows)
        col_index = 0
        existing_columns = copy.deepcopy(self.columns)
        for ci, col in enumerate(existing_columns):
            col["col_index"] = col_index
            new_cols.append(col)
            col_index += 1
            if col["kind"] == "struct":
                for fi,field in enumerate(col["field_type"].fields):
                    kind = self.__extract_kind__(field)
                    field_type = self.__extract_type__(field)
                    new_col = {"col_index": col_index, "name": f"*{field.name}", "kind": kind, "type": type(field.dataType).__name__, "field_type": field_type}
                    new_cols.append(new_col)

                    for row in new_rows:
                        struct_value = row["row"][ci]["value"]
                        row["row"].insert(col_index, struct_value["row"][fi])

                    col_index += 1

        self.columns = new_cols
        self.rows = new_rows
        self.__extract_column_names__()
        self.__extract_row_values__()


    @staticmethod
    def __extract_type__(field) -> StructType:
        # extract inner type from ArrayType, return field type otherwise
        if type(field.dataType) == ArrayType:
            return field.dataType.elementType

        return field.dataType


    @staticmethod
    def __extract_embedded_schema_tree__(fields: [StructField]) -> []:
        subfields = []
        for f in fields:
            subfield = {"name": f.name, "kind": DataFrameTable.__extract_kind__(f), "type": type(f.dataType).__name__}
            if subfield["kind"] == "array":
                embedded_subfields = DataFrameTable.__extract_embedded_array_to_tree__(f.dataType)
            elif subfield["kind"] == "struct":
                embedded_subfields = DataFrameTable.__extract_embedded_schema_tree__(f.dataType.fields)
            else:
                embedded_subfields = []
            subfield["subfields"] = embedded_subfields
            subfields.append(subfield)
        return subfields


    @staticmethod
    def __extract_embedded_array_to_tree__(subfield: DataType) -> []:
        if type(subfield) == StructType:
            subfields = DataFrameTable.__extract_embedded_schema_tree__(subfield.fields)
        elif type(subfield) == ArrayType:
            subfields = DataFrameTable.__extract_embedded_array_to_tree__(subfield.elementType)
        else:
            subfields = [{"name": "", "kind": "simple", "type": type(subfield).__name__, "subfields": []}]

        return subfields

    def __extract_schema_tree__(self) -> None:
        tree = []
        for col in self.columns:
            tree_field = {"name": col["name"], "kind": col["kind"], "type": col["type"]}
            if col["kind"] == "array":
                subfields = self.__extract_embedded_array_to_tree__(col["field_type"])
            elif col["kind"] == "struct":
                subfields = self.__extract_embedded_schema_tree__(col["field_type"])
            else:
                subfields = []
            tree_field["subfields"] = subfields
            tree.append(tree_field)

        self.schema_tree = tree


    def __extract_columns__(self) -> None:
        cols = []
        for i,field in enumerate(self._schema):
            field_type = self.__extract_type__(field)

            kind = self.__extract_kind__(field)
            cols.append({"col_index": i, "name": field.name, "kind": kind, "type": type(field.dataType).__name__, "field_type": field_type})

        self.columns = cols
        self.__extract_column_names__()
        self.__extract_schema_tree__()


    def __extract_column_names__(self) -> None:
        self.column_names = [c["name"] for c in self.columns]


    def __extract_rows__(self) -> None:
        assert len(self.columns) > 0  # ensure columns are calculated BEFORE rows

        rows = []
        for ri,data_row in enumerate(self._data):
            row=[]
            for fi, field in enumerate(data_row.__fields__):
                if self.columns[fi]["kind"] == "array":
                    # create internal schema as a single field
                    column = StructField(self.columns[fi]["name"], self.columns[fi]["field_type"])
                    # specify row schema in a form of name = value
                    values_as_row = list(map(lambda r: Row(**{self.columns[fi]["name"] : r}), data_row[field])) if data_row[field] is not None else None
                    value = DataFrameTable([column], values_as_row).rows
                    display_value = self.disp_value(data_row[field]) if data_row[field] is not None else "[]"
                elif self.columns[fi]["kind"] == "struct":
                    # extract internal schema as an array of fields
                    inner_schema = self.columns[fi]["field_type"].fields
                    # a value is just a single Row, so we must pack it as an array and then unpack it
                    value = DataFrameTable(inner_schema, [data_row[field]]).rows[0]
                    display_value = self.disp_value(data_row[field])
                else:
                    value = data_row[field]
                    display_value = self.disp_value(value)

                row.append({"display_value": display_value, "value": value})

            rows.append({"row_index": ri, "row": row})

        self.__set_rows__(rows)


    def __set_rows__(self, rows: []) -> None:
        self.rows = rows
        self.__extract_row_values__()


    def __extract_row_values__(self) -> None:
        # maybe it is not very readable but still it's one-liner
        self.row_values = [[c["display_value"][:DataFrameTable.TEXT_LEN] for c in r["row"]] for r in self.rows]


    def select(self, x: int, y: int) -> ({}, {}):
        return self.columns[x], self.rows[y]["row"][x]


    @staticmethod
    def disp_value(value: Any) -> str:
        disp_value = str(value)
        # remove single "Row(...)"
        disp_value = disp_value[4:-1] if disp_value.startswith("Row(") and disp_value.endswith(")") else disp_value
        # remove "Row" from array "[Row(...), ... , Row(...)]"
        disp_value = disp_value[4:-1].replace("), Row(", "), (") if disp_value.startswith("[Row(") and disp_value.endswith(")]") else disp_value
        return disp_value


def extract_embedded_table(tab: DataFrameTable, x: int, y: int, expand_structs: bool = False) -> DataFrameTable | None:
    # check for kind=array but field_type=StructType - this means that we want to drill down from array to structs
    column, cell = tab.select(x,y)
    kind = column["kind"]
    if kind == "struct": # or type(column["field_type"]) == StructType:
        columns = copy.deepcopy(column["field_type"].fields)
        rows = copy.deepcopy([cell["value"]])
        new_tab = DataFrameTable(columns, data= [], transformed_data=rows, expand_structs= expand_structs)
        return new_tab

    # other case for array
    if kind=="array":
        col_name = column["name"][1:] if column["name"].startswith("*") else column["name"]
        columns = copy.deepcopy([StructField(col_name, column["field_type"])])
        rows = copy.deepcopy(cell["value"])
        new_tab = DataFrameTable(columns, data= [], transformed_data=rows, expand_structs= expand_structs)
        return new_tab

    return None
